#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_code.h"
#include "asm_code_riscv32.h"

static int asm_code_riscv32_get_word_size(void)
{
  return 4;
}

static int asm_code_riscv32_get_pointer_size(void)
{
  return asm_code_riscv32_get_word_size();
}

static int asm_code_riscv32_funccall_args_reg_number(void)
{
  return 8; /* a0/a1/a2/a3/a4/a5/a6/a7 (See funccall_regs[]) */
}

static int asm_code_riscv32_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_riscv32_tmp_reg_number(void)
{
  return 2; /* s0/s1 (See tmp_regs[]) */
}

static int asm_code_riscv32_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_riscv32_function_register_number(void)
{
  return 2; /* s0/s1 (See asm_code_riscv32_function_register_save() */
}

static int asm_code_riscv32_function_saveparam_number(void)
{
  return 1; /* ra (See asm_code_riscv32_function_start() */
}

static int asm_code_riscv32_stack_align_size(void)
{
  return 16;
}

static int asm_code_riscv32_stack_correct_size(void)
{
  return 0;
}

static char *load_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "lb";
    case  2: return "lh";
    default: break;
    }
  }
  return "lw";
}

static char *store_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "sb";
    case  2: return "sh";
    default: break;
    }
  }
  return "sw";
}

static void asm_code_riscv32_get_value(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tlui\ta0, 0x%lx\n", (value >> 20) & 0xfffff);
  ASM_CODE_OUT(out, "\taddi\ta0, a0, 0x%lx\n", value & 0xfff);
}

static void asm_code_riscv32_get_value_r1(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tlui\ta1, 0x%lx\n", (value >> 20) & 0xfffff);
  ASM_CODE_OUT(out, "\taddi\ta1, a1, 0x%lx\n", value & 0xfff);
}

static void asm_code_riscv32_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmv\ta0, sp\n");
}

static void asm_code_riscv32_get_address(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tlui\ta0, %%hi(%s)\n", label);
  ASM_CODE_OUT(out, "\taddi\ta0, a0, %%lo(%s)\n", label);
}

static void asm_code_riscv32_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\taddi\ta0, a0, 0x%x\n", offset);
}

static void asm_code_riscv32_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmv\ta0, a1\n");
}

static void asm_code_riscv32_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmv\ta1, a0\n");
}

static void asm_code_riscv32_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta0, 0x%x(a1)\n", load_by_size(model), offset);
}

static void asm_code_riscv32_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta0, 0x%x(a1)\n", store_by_size(model), offset);
}

static void asm_code_riscv32_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta0, 0x%x(sp)\n", load_by_size(model), offset);
}

static void asm_code_riscv32_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta0, 0x%x(sp)\n", store_by_size(model), offset);
}

static void asm_code_riscv32_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta1, 0x%x(sp)\n", load_by_size(model), offset);
}

static void asm_code_riscv32_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\t%s\ta1, 0x%x(sp)\n", store_by_size(model), offset);
}

static void asm_code_riscv32_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\taddi\tsp, sp, -0x%x\n", size);
}

static void asm_code_riscv32_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\taddi\tsp, sp, 0x%x\n", size);
}

static char *funccall_regs[] = {
  "a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"
};

static void asm_code_riscv32_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\ta0, %s\n", funccall_regs[number]);
}

static void asm_code_riscv32_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\t%s, a0\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "s0", "s1",
};

static void asm_code_riscv32_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\ta0, %s\n", tmp_regs[number]);
}

static void asm_code_riscv32_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\t%s, a0\n", tmp_regs[number]);
}

static void asm_code_riscv32_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\ta1, %s\n", tmp_regs[number]);
}

static void asm_code_riscv32_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmv\t%s, a1\n", tmp_regs[number]);
}

static int asm_code_riscv32_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tslli\ta0, a0, 24\n");
  ASM_CODE_OUT(out, "\tsrai\ta0, a0, 24\n");
  return 0;
}

static int asm_code_riscv32_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tandi\ta0, a0, 0xff\n");
  return 0;
}

static int asm_code_riscv32_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tslli\ta0, a0, 16\n");
  ASM_CODE_OUT(out, "\tsrai\ta0, a0, 16\n");
  return 0;
}

static int asm_code_riscv32_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tslli\ta0, a0, 16\n");
  ASM_CODE_OUT(out, "\tsrli\ta0, a0, 16\n");
  return 0;
}

static int asm_code_riscv32_sign_extension_int(FILE *out)
{
  return -1;
}

static int asm_code_riscv32_sign_extension_uint(FILE *out)
{
  return -1;
}

static int asm_code_riscv32_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tseqz\ta0, a0\n");
  return 0;
}

static int asm_code_riscv32_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tneg\ta0, a0\n");
  return 0;
}

static int asm_code_riscv32_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_riscv32_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadd\ta0, a0, a1\n");
}

static void asm_code_riscv32_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsub\ta0, a0, a1\n");
}

static void asm_code_riscv32_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tand\ta0, a0, a1\n");
}

static void asm_code_riscv32_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tor\ta0, a0, a1\n");
}

static void asm_code_riscv32_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\txor\ta0, a0, a1\n");
}

static int asm_code_riscv32_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmul\ta0, a0, a1\n");
  return 0;
}

static int asm_code_riscv32_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tdiv\ta0, a0, a1\n");
  } else {
    ASM_CODE_OUT(out, "\tdivu\ta0, a0, a1\n");
  }
  return 0;
}

static int asm_code_riscv32_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\trem\ta0, a0, a1\n");
  } else {
    ASM_CODE_OUT(out, "\tremu\ta0, a0, a1\n");
  }
  return 0;
}

static int asm_code_riscv32_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsll\ta0, a0, a1\n");
  return 0;
}

static int asm_code_riscv32_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsra\ta0, a0, a1\n");
  return 0;
}

static int asm_code_riscv32_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsrl\ta0, a0, a1\n");
  return 0;
}

static int asm_code_riscv32_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_riscv32_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tj\t%s\n", label);
}

static int asm_code_riscv32_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tbeqz\ta0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tbnez\ta0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbeq\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbne\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tblt\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tblt\ta1, a0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tble\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tble\ta1, a0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbltu\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbltu\ta1, a0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbleu\ta0, a1, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tbleu\ta1, a0, %s\n", label);
  return 0;
}

static int asm_code_riscv32_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static int asm_code_riscv32_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tjal\t%s\n", label);
  return 0;
}

static int asm_code_riscv32_function_call_set(FILE *out)
{
  ASM_CODE_OUT(out, "\tmv\tra, a0\n");
  return 0;
}

static int asm_code_riscv32_function_call_pointer(FILE *out)
{
  ASM_CODE_OUT(out, "\tjalr\tra\n");
  return 0;
}

static void asm_code_riscv32_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\taddi\tsp, sp, -4\n");
  ASM_CODE_OUT(out, "\tsw\tra, 0(sp)\n");
}

static void asm_code_riscv32_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tlw\tra, 0(sp)\n");
  ASM_CODE_OUT(out, "\taddi\tsp, sp, 4\n");
  ASM_CODE_OUT(out, "\tret\n\n");
}

static void asm_code_riscv32_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\taddi\tsp, sp, -8\n");
  ASM_CODE_OUT(out, "\tsw\ts0, 0(sp)\n");
  ASM_CODE_OUT(out, "\tsw\ts1, 4(sp)\n");
}

static void asm_code_riscv32_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tlw\ts0, 0(sp)\n");
  ASM_CODE_OUT(out, "\tlw\ts1, 4(sp)\n");
  ASM_CODE_OUT(out, "\taddi\tsp, sp, 8\n");
}

struct asm_code_callback asm_code_callback_riscv32 = {
  asm_code_riscv32_get_word_size,
  asm_code_riscv32_get_pointer_size,

  asm_code_riscv32_funccall_args_reg_number,
  asm_code_riscv32_funccall_args_stack_number,
  asm_code_riscv32_tmp_reg_number,
  asm_code_riscv32_tmp_stack_number,
  asm_code_riscv32_function_register_number,
  asm_code_riscv32_function_saveparam_number,

  asm_code_riscv32_stack_align_size,
  asm_code_riscv32_stack_correct_size,

  asm_code_riscv32_get_value,
  asm_code_riscv32_get_value_r1,
  asm_code_riscv32_get_address_stack,
  asm_code_riscv32_get_address,
  asm_code_riscv32_add_address,
  asm_code_riscv32_get_r1,
  asm_code_riscv32_set_r1,

  asm_code_riscv32_memory_load,
  asm_code_riscv32_memory_store,
  asm_code_riscv32_stack_load,
  asm_code_riscv32_stack_store,
  asm_code_riscv32_stack_load_r1,
  asm_code_riscv32_stack_store_r1,
  asm_code_riscv32_stack_expand,
  asm_code_riscv32_stack_reduce,

  asm_code_riscv32_funccall_reg_load,
  asm_code_riscv32_funccall_reg_store,
  asm_code_riscv32_tmp_reg_load,
  asm_code_riscv32_tmp_reg_save,
  asm_code_riscv32_tmp_reg_load_r1,
  asm_code_riscv32_tmp_reg_save_r1,

  asm_code_riscv32_sign_extension_char,
  asm_code_riscv32_sign_extension_uchar,
  asm_code_riscv32_sign_extension_short,
  asm_code_riscv32_sign_extension_ushort,
  asm_code_riscv32_sign_extension_int,
  asm_code_riscv32_sign_extension_uint,

  asm_code_riscv32_calc_inv,
  asm_code_riscv32_calc_minus,

  asm_code_riscv32_calc_op1,

  asm_code_riscv32_calc_add,
  asm_code_riscv32_calc_sub,
  asm_code_riscv32_calc_and,
  asm_code_riscv32_calc_or,
  asm_code_riscv32_calc_xor,

  asm_code_riscv32_calc_mul,
  asm_code_riscv32_calc_div,
  asm_code_riscv32_calc_mod,
  asm_code_riscv32_calc_llshift,
  asm_code_riscv32_calc_rashift,
  asm_code_riscv32_calc_rlshift,

  asm_code_riscv32_calc_op2,

  asm_code_riscv32_branch,
  asm_code_riscv32_branch_zero,
  asm_code_riscv32_branch_nzero,

  asm_code_riscv32_branch_cmp_eq,
  asm_code_riscv32_branch_cmp_ne,

  asm_code_riscv32_branch_cmp_lt,
  asm_code_riscv32_branch_cmp_gt,
  asm_code_riscv32_branch_cmp_le,
  asm_code_riscv32_branch_cmp_ge,
  asm_code_riscv32_branch_cmp_ult,
  asm_code_riscv32_branch_cmp_ugt,
  asm_code_riscv32_branch_cmp_ule,
  asm_code_riscv32_branch_cmp_uge,

  asm_code_riscv32_branch_cmp,

  asm_code_riscv32_function_call,
  asm_code_riscv32_function_call_set,
  asm_code_riscv32_function_call_pointer,

  asm_code_riscv32_function_start,
  asm_code_riscv32_function_end,
  asm_code_riscv32_function_register_save,
  asm_code_riscv32_function_register_load,
};
